import json
from typing import Any, Dict

from .utils import DEFAULT_HEIGHT, stoys_ui_component_html

_default_aggsum_result_ui_config = dict(
    previousReleaseDataIsShown=False,
    # disabledColumns=[],
    pagination=False,
    height=DEFAULT_HEIGHT,
    smallSize=True,
)


def aggsum_result_to_html(aggsum_result, previous_aggsum_result=None, **kwargs: Dict[str, Any]) -> str:  # noqa: ANN001
    payload = dict(
        data=dict(
            current=_aggsum_result_to_dict(aggsum_result),
            previous=_aggsum_result_to_dict(previous_aggsum_result),
        ),
        config={**_default_aggsum_result_ui_config, **kwargs},
    )
    return stoys_ui_component_html("Metrics", payload)


def _aggsum_result_to_dict(aggsum_result) -> Any:  # noqa: ANN001
    if aggsum_result is None:
        return None
    else:
        rtn = aggsum_result.aggsum_info.referenced_table_names
        return dict(
            table_name=rtn[0] if rtn else None,
            key_columns=aggsum_result.aggsum_info.key_column_names,
            data=json.loads(aggsum_result.data_json),
        )



from .aggsum import AggSum
from .aggsum_config import AggSumConfig
from .aggsum_result import AggSumInfo, AggSumResult

__all__ = [
    "AggSum",
    "AggSumConfig",
    "AggSumInfo",
    "AggSumResult",
]



from enum import auto

from ...utils.python import AutoNameEnum


class DqJoinType(AutoNameEnum):
    UNDEFINED = auto()
    LEFT = auto()
    RIGHT = auto()
    INNER = auto()
    FULL = auto()
    CROSS = auto()



from typing import Any, Type

import serde.json


class Json:
    @staticmethod
    def to_json(value: Any) -> str:
        return serde.json.to_json(value)

    @staticmethod
    def from_json(cls: Type, json: str) -> str:
        return serde.json.from_json(cls, json)



from typing import ClassVar

from py4j.java_gateway import JavaClass, JavaObject, get_java_class

from ..utils.jvm import JvmPackage
from ..utils.python import classproperty

_jvm_scala = JvmPackage("io.stoys.scala")


class Jackson:
    _json: ClassVar[JavaObject] = None

    @classproperty
    def json(cls) -> JavaClass:  # noqa: B902,N805
        if cls._json is None:
            cls._json = _jvm_scala.Jackson.json()
        return cls._json

    @staticmethod
    def toJson(java_object: JavaObject) -> str:
        return Jackson.json.writeValueAsString(java_object)

    @staticmethod
    def fromJson(java_class: JavaClass, json: str) -> JavaObject:
        return Jackson.json.readValue(json, get_java_class(java_class))



from typing import Dict, List, Optional

from ...utils.case_class import CaseClassMirror, case_class_mirror


@case_class_mirror("io.stoys.spark.dp.DpPmfBucket")
class DpPmfBucket(CaseClassMirror):
    low: float
    high: float
    count: int


@case_class_mirror("io.stoys.spark.dp.DpItem")
class DpItem(CaseClassMirror):
    item: str
    count: int


@case_class_mirror("io.stoys.spark.dp.DpColumn")
class DpColumn(CaseClassMirror):
    name: str
    data_type: str
    data_type_json: str
    nullable: bool
    enum_values: List[str]
    format: Optional[str]
    count: int
    count_empty: Optional[int]
    count_nulls: Optional[int]
    count_unique: Optional[int]
    count_zeros: Optional[int]
    max_length: Optional[int]
    min: Optional[str]
    max: Optional[str]
    mean: Optional[float]
    pmf: List[DpPmfBucket]
    items: List[DpItem]
    extras: Dict[str, str]


@case_class_mirror("io.stoys.spark.dp.DpTable")
class DpTable(CaseClassMirror):
    rows: int


@case_class_mirror("io.stoys.spark.dp.DpResult")
class DpResult(CaseClassMirror):
    table: DpTable
    columns: List[DpColumn]

    def _repr_html_(self) -> str:
        from ...ui.dp_ui import dp_result_to_html

        return dp_result_to_html(self)



from typing import List

from ...utils.case_class import CaseClassMirror, case_class_mirror


@case_class_mirror("io.stoys.spark.aggsum.AggSumInfo")
class AggSumInfo(CaseClassMirror):
    key_column_names: List[str]
    value_column_names: List[str]
    referenced_table_names: List[str]


@case_class_mirror("io.stoys.spark.aggsum.AggSumResult")
class AggSumResult(CaseClassMirror):
    aggsum_info: AggSumInfo
    data_json: str  # df.to_json(orient='records')

    def _repr_html_(self) -> str:
        from ...ui.aggsum_ui import aggsum_result_to_html

        return aggsum_result_to_html(self)



from typing import Optional

from ..utils.jvm import JvmPackage
from ..utils.py4j import Py4j

_jvm_spark = JvmPackage("io.stoys.spark")


class TableName:
    def __init__(self, entity_name: str, logical_name: Optional[str] = None) -> None:
        self._table_name = _jvm_spark.TableName(entity_name, Py4j.toOption(logical_name), None)  # TODO: tag

    def full_table_name(self) -> str:
        self._table_name.fullTableName()



from typing import Any, Dict

from .utils import stoys_ui_component_html

_default_dq_join_result_ui_config = dict(
    # mode="row",
    # mode="column",
    smallSize=True,
    # selectedRules=[],
    pagination=False,
    # statisticsTableProps={},
    # qualityTableProps={},
)


def dq_join_result_to_html(*dq_join_result, **kwargs: Dict[str, Any]) -> str:  # noqa: ANN002
    payload = dict(
        data=list(dq_join_result),
        config={**_default_dq_join_result_ui_config, **kwargs},
    )
    return stoys_ui_component_html("JoinRates", payload)



from .arbitrary import Arbitrary
from .jackson import Jackson
from .reflection import Reflection
from .strings import Strings

__all__ = [
    "Arbitrary",
    "Jackson",
    "Reflection",
    "Strings",
]



from __future__ import annotations

from typing import Any, Dict, List, Optional

from pyspark.sql import DataFrame, SparkSession

from ..utils.jvm import JvmPackage
from ..utils.py4j import Py4j
from .reshape import ReshapeConfig
from .spark_io_config import SparkIOConfig
from .table_name import TableName

_jvm_spark = JvmPackage("io.stoys.spark")


class SparkIO:
    def __init__(self, spark_session: SparkSession, config: SparkIOConfig) -> None:
        self._sql_ctx = spark_session._jsc
        self._spark_io = _jvm_spark.SparkIO(spark_session._jsparkSession, config.to_jvm())

    def df(self, entity_name: str, logical_name: Optional[str] = None) -> DataFrame:
        table_name = TableName(entity_name, logical_name)
        return DataFrame(self._spark_io.df(table_name), self._sql_ctx)

    def write(
        self,
        df: DataFrame,
        entity_name: str,
        logical_name: Optional[str] = None,
        format: Optional[str] = None,
        write_mode: Optional[str] = None,
        options: Optional[Dict[str, str]] = None,
    ) -> None:
        table_name = TableName(entity_name, logical_name)
        self._spark_io.writeDF(
            df._jdf,
            table_name.fullTableName(),
            Py4j.toOption(format),
            Py4j.toOption(write_mode),
            Py4j.toMap(options or {}),
        )

    def close(self) -> Any:
        return self._spark_io.close()


class SparkIOContext:
    def __init__(
        self,
        spark_session: SparkSession,
        input_paths: List[str],
        output_path: Optional[str],
        write_format: Optional[str] = None,
        write_mode: Optional[str] = None,
        write_options: Optional[Dict[str, str]] = None,
    ) -> None:
        self._spark_session = spark_session
        self._config = SparkIOConfig(
            input_paths,
            ReshapeConfig.default,
            True,
            output_path,
            write_format,
            write_mode,
            write_options or {},
        )

    def __enter__(self) -> SparkIO:
        self._spark_io = SparkIO(self._spark_session, self._config)
        return self._spark_io

    def __exit__(self, exc_type, exc_val, exc_tb) -> Any:  # noqa: ANN001
        return self._spark_io.close()



from __future__ import annotations

from typing import Dict, List

from py4j.java_gateway import JavaObject
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import SparkSession, SQLContext

from ...utils.jvm import JvmPackage
from ...utils.py4j import Py4j
from .dq_config import DqConfig
from .dq_field import DqField
from .dq_result import DqResult
from .dq_rule import DqRule

_jvm_dq = JvmPackage("io.stoys.spark.dq")


class Dq:
    def __init__(self, sql_ctx: SQLContext, java_dq: JavaObject) -> None:
        self._sql_ctx = sql_ctx
        self._java_dq = java_dq

    def config(self, config: DqConfig) -> Dq:
        if config:
            self._java_dq.config(config.to_jvm())
        return self

    def fields(self, fields: List[DqField]) -> Dq:
        if fields:
            self._java_dq.fields(Py4j.toSeq([f.to_jvm() for f in fields]))
        return self

    def metadata(self, metadata: Dict[str, str]) -> Dq:
        if metadata:
            self._java_dq.metadata(Py4j.toMap(metadata))
        return self

    def primaryKeyFieldNames(self, primary_key_field_names: List[str]) -> Dq:
        if primary_key_field_names:
            self._java_dq.primaryKeyFieldNames(Py4j.toSeq(primary_key_field_names))
        return self

    def rules(self, rules: List[DqRule]) -> Dq:
        if rules:
            self._java_dq.rules(Py4j.toSeq([r.to_jvm() for r in rules]))
        return self

    def computeDqResult(self) -> SparkDataFrame:
        return SparkDataFrame(self._java_dq.computeDqResult(), self._sql_ctx)

    def computeDqViolationPerRow(self) -> SparkDataFrame:
        return SparkDataFrame(self._java_dq.computeDqViolationPerRow(), self._sql_ctx)

    def selectFailingRows(self) -> SparkDataFrame:
        return SparkDataFrame(self._java_dq.selectFailingRows(), self._sql_ctx)

    def selectPassingRows(self) -> SparkDataFrame:
        return SparkDataFrame(self._java_dq.selectPassingRows(), self._sql_ctx)

    def dqResult(self) -> DqResult:
        return DqResult.from_jvm(self._java_dq.computeDqResult().first())

    @staticmethod
    def fromDataFrame(sdf: SparkDataFrame) -> Dq:
        return Dq(sdf.sql_ctx, _jvm_dq.Dq.fromDataFrame(sdf._jdf))

    # @staticmethod
    # def fromDataset(sdf: SparkDataFrame) -> Dq:
    #     return Dq(sdf.sql_ctx, jvm_dq.Dq.fromDataset(sdf._jdf))

    @staticmethod
    def fromDqSql(spark_session: SparkSession, dq_sql: str) -> Dq:
        sql_ctx = SQLContext(spark_session.sparkContext)
        return Dq(sql_ctx, _jvm_dq.Dq.fromDqSql(spark_session._jsparkSession, dq_sql))

    @staticmethod
    def fromFileInputPath(spark_session: SparkSession, input_path: str) -> Dq:
        sql_ctx = SQLContext(spark_session.sparkContext)
        return Dq(
            sql_ctx,
            _jvm_dq.Dq.fromFileInputPath(spark_session._jsparkSession, input_path),
        )

    @staticmethod
    def fromTableName(spark_session: SparkSession, table_name: str) -> Dq:
        sql_ctx = SQLContext(spark_session.sparkContext)
        return Dq(sql_ctx, _jvm_dq.Dq.fromTableName(spark_session._jsparkSession, table_name))



from typing import List, Optional

from ...utils.case_class import CaseClassMirror, case_class_mirror


@case_class_mirror("io.stoys.spark.dq.DqField")
class DqField(CaseClassMirror):
    name: str
    data_type_json: str
    nullable: bool
    enum_values: List[str]
    format: Optional[str]
    regexp: Optional[str]



import pathlib
import urllib.request
from typing import List

from ..config import (
    STOYS_LIB_CLASSIFIER,
    STOYS_LIB_GROUP_ID,
    STOYS_LIB_MODULES,
    STOYS_LIB_SCALA_COMPAT_VERSION,
    STOYS_LIB_VERSION,
    STOYS_UI_VERSION,
)
from .common import embedded_dependencies_dir


def _download_file(url: str, path: pathlib.Path) -> pathlib.Path:
    if not path.exists():
        path.parent.mkdir(parents=True, exist_ok=True)
        if url.lower().startswith("https"):
            urllib.request.urlretrieve(url, path)  # noqa: S310
        else:
            raise ValueError from None
    return path


def download_stoys_lib_jar(target_dir: pathlib.Path, module: str) -> List[pathlib.Path]:
    base_url = "https://repo1.maven.org/maven2"
    group_dir = STOYS_LIB_GROUP_ID.replace(".", "/")
    artifact_dir = f"{module}_{STOYS_LIB_SCALA_COMPAT_VERSION}/{STOYS_LIB_VERSION}"
    file_name = f"{module}_{STOYS_LIB_SCALA_COMPAT_VERSION}-{STOYS_LIB_VERSION}-{STOYS_LIB_CLASSIFIER}.jar"
    url = f"{base_url}/{group_dir}/{artifact_dir}/{file_name}"
    return _download_file(url, target_dir / file_name)


def download_stoys_ui_js(target_dir: pathlib.Path) -> pathlib.Path:
    base_url = "https://github.com/stoys-io/stoys-ui/releases/download"
    file_name = "stoys-ui.js"
    url = f"{base_url}/{STOYS_UI_VERSION}/{file_name}"
    return _download_file(url, target_dir / file_name)


def download_dependencies(
    dependencies_dir: pathlib.Path = embedded_dependencies_dir,
) -> List[pathlib.Path]:
    lib_deps = [download_stoys_lib_jar(dependencies_dir, m) for m in STOYS_LIB_MODULES]
    ui_deps = [download_stoys_ui_js(dependencies_dir)]
    return lib_deps + ui_deps


if __name__ == "__main__":
    download_dependencies()



from __future__ import annotations

from ...utils.case_class import CaseClassMirror, case_class_mirror
from ...utils.python import classproperty


@case_class_mirror("io.stoys.spark.dq.DqConfig")
class DqConfig(CaseClassMirror):
    sample_rows: bool
    max_rows_per_rule: int
    max_rows: int
    report_extra_columns: bool

    @classproperty
    def default(cls) -> DqConfig:  # noqa: B902, N805
        return cls.from_jvm(cls.java_class.default())



from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql.types import DataType

from ..utils.data_types import DataTypes
from ..utils.jvm import JvmPackage
from .reshape_config import ReshapeConfig

_jvm_spark = JvmPackage("io.stoys.spark")


class Reshape:
    @staticmethod
    def reshapeToDF(sdf: SparkDataFrame, target_schema: DataType, config: ReshapeConfig = None) -> SparkDataFrame:
        jvm_target_schema = DataTypes.to_jvm(target_schema)
        jvm_config = (config or ReshapeConfig.default).to_jvm()
        reshaped_jdf = _jvm_spark.Reshape.reshapeToDF(sdf._jdf, jvm_target_schema, jvm_config)
        return SparkDataFrame(reshaped_jdf, sdf.sql_ctx)



from __future__ import annotations

from typing import Any, Optional

from .python import classproperty


def Option(value: Any) -> Optional[Any]:
    return value


def Some(value: Any) -> Optional[Any]:
    return value


class Seq(list):
    def __init__(self, *args: Any) -> None:
        list.__init__(self, list(args))

    @classproperty
    def empty(cls) -> Seq[Any]:  # noqa: B902,N805
        return Seq()


class Map(dict):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # if len(args) > 0 and isinstance(args[0], tuple):
        # dict.__init__(self, *args, **kwargs)
        dict.__init__(self, {**dict(list(args)), **dict(kwargs)})

    @classproperty
    def empty(cls) -> Map[Any]:  # noqa: B902,N805
        return Map()



from typing import Any, Dict, List

from py4j.java_gateway import JavaObject

from .jvm import JvmPackage
from .python import classproperty

_jvm_utils = JvmPackage("io.stoys.utils")


class Py4j:
    @classproperty
    def emptyOption(cls) -> JavaObject:  # noqa: B902,N805
        return _jvm_utils.Py4j.emptyOption()

    @classproperty
    def emptySeq(cls) -> JavaObject:  # noqa: B902,N805
        return _jvm_utils.Py4j.emptySeq()

    @classproperty
    def emptyMap(cls) -> JavaObject:  # noqa: B902,N805
        return _jvm_utils.Py4j.emptyMap()

    @staticmethod
    def toArray(values: List[Any]) -> JavaObject:
        return _jvm_utils.Py4j.toArray(values)

    @staticmethod
    def toList(values: List[Any]) -> JavaObject:
        return _jvm_utils.Py4j.toList(values)

    @staticmethod
    def toMap(key_values: Dict[Any, Any]) -> JavaObject:
        return _jvm_utils.Py4j.toMap(key_values)

    @staticmethod
    def toOption(value: Any) -> JavaObject:
        return _jvm_utils.Py4j.toOption(value)

    @staticmethod
    def toSeq(values: List[Any]) -> JavaObject:
        return _jvm_utils.Py4j.toSeq(values)

    @staticmethod
    def toSet(values: List[Any]) -> JavaObject:
        return _jvm_utils.Py4j.toSet(values)



from typing import Dict, List

from ...utils.case_class import CaseClassMirror, case_class_mirror
from .dq_rule import DqRule


@case_class_mirror("io.stoys.spark.dq.DqColumn")
class DqColumn(CaseClassMirror):
    name: str


@case_class_mirror("io.stoys.spark.dq.DqTableStatistic")
class DqTableStatistic(CaseClassMirror):
    rows: int
    violations: int


@case_class_mirror("io.stoys.spark.dq.DqRuleStatistics")
class DqRuleStatistics(CaseClassMirror):
    rule_name: str
    violations: int


@case_class_mirror("io.stoys.spark.dq.DqColumnStatistics")
class DqColumnStatistics(CaseClassMirror):
    column_name: str
    violations: int


@case_class_mirror("io.stoys.spark.dq.DqStatistics")
class DqStatistics(CaseClassMirror):
    table: DqTableStatistic
    column: List[DqColumnStatistics]
    rule: List[DqRuleStatistics]


@case_class_mirror("io.stoys.spark.dq.DqRowSample")
class DqRowSample(CaseClassMirror):
    row: List[str]
    violated_rule_names: List[str]


@case_class_mirror("io.stoys.spark.dq.DqResult")
class DqResult(CaseClassMirror):
    columns: List[DqColumn]
    rules: List[DqRule]
    statistics: DqStatistics
    row_sample: List[DqRowSample]
    metadata: Dict[str, str]

    def _repr_html_(self) -> str:
        from ...ui.dq_result_ui import dq_result_to_html

        return dq_result_to_html(self)



from ..utils.jvm import JvmPackage

_jvm_scala = JvmPackage("io.stoys.scala")


class Strings:
    @staticmethod
    def to_snake_case(value: str) -> str:
        return _jvm_scala.Strings.toSnakeCase(value)

    @staticmethod
    def trim(value: str) -> str:
        return _jvm_scala.Strings.trim(value)

    @staticmethod
    def to_word_characters(value: str) -> str:
        return _jvm_scala.Strings.toWordCharacters(value)

    @staticmethod
    def to_word_characters_collapsing(value: str) -> str:
        return _jvm_scala.Strings.toWordCharactersCollapsing(value)



from __future__ import annotations

from typing import Optional

from ...utils.case_class import CaseClassMirror, case_class_mirror
from ...utils.python import classproperty
from .dp_type_inference_config import DpTypeInferenceConfig


@case_class_mirror("io.stoys.spark.dp.DpConfig")
class DpConfig(CaseClassMirror):
    max_item_length: int
    items: int
    pmf_buckets: int
    time_zone_id: Optional[str]
    infer_types_from_strings: bool
    type_inference_config: DpTypeInferenceConfig

    @classproperty
    def default(cls) -> DpConfig:  # noqa: B902, N805
        return cls.from_jvm(cls.java_class.default())



from __future__ import annotations

from typing import Optional

from ...utils.case_class import CaseClassMirror, case_class_mirror
from ...utils.python import classproperty


@case_class_mirror("io.stoys.spark.aggsum.AggSumConfig")
class AggSumConfig(CaseClassMirror):
    limit: Optional[bool]

    @classproperty
    def default(cls) -> AggSumConfig:  # noqa: B902, N805
        return cls.from_jvm(cls.java_class.default())



from py4j.java_gateway import JavaObject, is_instance_of
from pyspark.sql.types import DataType, StructType

from .jvm import JvmPackage

_jvm_spark_sql_types = JvmPackage("org.apache.spark.sql.types")


class DataTypes:
    @staticmethod
    def to_jvm(data_type: DataType) -> JavaObject:
        return _jvm_spark_sql_types.DataType.fromJson(data_type.json())

    @staticmethod
    def from_jvm(jvm_data_type: JavaObject) -> DataType:
        if is_instance_of(jvm_data_type, "org.apache.spark.sql.types.StructType"):
            return StructType.fromJson(jvm_data_type.json())
        else:
            empty_metadata = _jvm_spark_sql_types.Metadata.empty()
            jvm_struct_field = _jvm_spark_sql_types.StructField("dummy", jvm_data_type, True, empty_metadata)
            jvm_struct_type = _jvm_spark_sql_types.StructType([jvm_struct_field])
            struct_type = StructType.fromJson(jvm_struct_type.json())
            return struct_type.fields[0].dataType



from __future__ import annotations

from enum import auto
from typing import Optional

from ..utils.case_class import CaseClassMirror, case_class_mirror
from ..utils.python import AutoNameEnum, classproperty


class ReshapeConflictResolution(AutoNameEnum):
    UNDEFINED = auto()
    ERROR = auto()
    FIRST = auto()
    LAST = auto()


class ReshapeSortOrder(AutoNameEnum):
    UNDEFINED = auto()
    ALPHABETICAL = auto()
    SOURCE = auto()
    TARGET = auto()


class ReshapeFieldMatchingStrategy(AutoNameEnum):
    UNDEFINED = auto()
    NAME_DEFAULT = auto()
    NAME_EXACT = auto()
    NAME_NORMALIZED = auto()
    INDEX = auto()


@case_class_mirror("io.stoys.spark.ReshapeConfig")
class ReshapeConfig(CaseClassMirror):
    coerce_types: bool
    conflict_resolution: ReshapeConflictResolution
    drop_extra_columns: bool
    fail_on_extra_column: bool
    fail_on_ignoring_nullability: bool
    fill_default_values: bool
    fill_missing_nulls: bool
    field_matching_strategy: ReshapeFieldMatchingStrategy
    sort_order: ReshapeSortOrder
    date_format: Optional[str]
    timestamp_format: Optional[str]

    @classproperty
    def spark(cls) -> ReshapeConfig:  # noqa: B902,N805
        return cls.from_jvm(cls.java_class.spark())

    @classproperty
    def safe(cls) -> ReshapeConfig:  # noqa: B902,N805
        return cls.from_jvm(cls.java_class.safe())

    @classproperty
    def default(cls) -> ReshapeConfig:  # noqa: B902,N805
        return cls.from_jvm(cls.java_class.default())

    @classproperty
    def dangerous(cls) -> ReshapeConfig:  # noqa: B902,N805
        return cls.from_jvm(cls.java_class.dangerous())

    @classproperty
    def notebook(cls) -> ReshapeConfig:  # noqa: B902,N805
        return cls.from_jvm(cls.java_class.notebook())



from __future__ import annotations

from py4j.java_gateway import JavaObject
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import SQLContext

from ...utils.jvm import JvmPackage
from .dp_config import DpConfig
from .dp_result import DpResult

_jvm_dp = JvmPackage("io.stoys.spark.dp")


class Dp:
    def __init__(self, sql_ctx: SQLContext, java_dp: JavaObject) -> None:
        self._sql_ctx = sql_ctx
        self._java_dp = java_dp

    def config(self, config: DpConfig) -> Dp:
        if config:
            self._java_dp.config(config.to_jvm())
        return self

    def computeDpResult(self) -> SparkDataFrame:
        return SparkDataFrame(self._java_dp.computeDpResult(), self._sql_ctx)

    def dpResult(self) -> DpResult:
        return DpResult.from_jvm(self._java_dp.computeDpResult().first())

    @staticmethod
    def fromDataFrame(sdf: SparkDataFrame) -> Dp:
        return Dp(sdf.sql_ctx, _jvm_dp.Dp.fromDataFrame(sdf._jdf))

    # @staticmethod
    # def fromDataset(sdf: SparkDataFrame) -> Dp:
    #     return Dp(sdf.sql_ctx, _jvm_dp.Dp.fromDataset(sdf._jdf))



from dataclasses import dataclass
from typing import List


@dataclass
class PmfBucket:
    low: float
    high: float
    count: int


@dataclass
class Pmf:
    pmf: List[PmfBucket]

    def _repr_html_(self) -> str:
        from .ui.pmf_ui import pmf_to_html

        return pmf_to_html(self)



from __future__ import annotations

from typing import List, Optional, Union

from py4j.java_gateway import JavaClass, JavaPackage, JVMView
from pyspark import SparkContext
from pyspark.sql import SparkSession

from .spark_session import get_or_create_spark_session


class JvmPackage:
    def __init__(
        self,
        package_name: Optional[str] = None,
        *,
        context: Optional[Union[JVMView, SparkContext, SparkSession]] = None,
    ) -> None:
        self._package_name = package_name
        self._class_cache = {}
        self._jvm_view = None
        if context is not None:
            if isinstance(context, JVMView):
                self._jvm_view = context
            elif isinstance(context, SparkSession):
                self._jvm_view = context._jvm
            elif isinstance(context, SparkContext):
                self._jvm_view = context._jvm
            else:
                raise Exception(f"Unsupported context type ({type(context)}).")
            self._check_stoys_installed()

    def __dir__(self) -> List[str]:
        return [*self.__dict__, *self._class_cache]

    def __getattr__(self, name: str) -> Union[JavaClass, JvmPackage]:
        if name not in self._class_cache:
            full_name = name if self._package_name is None else f"{self._package_name}.{name}"
            java_class_or_package = getattr(self._jvm, full_name)
            if isinstance(java_class_or_package, JavaPackage):
                self._class_cache[name] = JvmPackage(java_class_or_package.fqn)
            else:
                self._class_cache[name] = java_class_or_package
        return self._class_cache[name]

    def __call__(self, name: str) -> Union[JavaClass, JvmPackage]:
        return self.__getattr__(name)

    @property
    def _jvm(self) -> JVMView:
        if self._jvm_view is None:
            self._jvm_view = get_or_create_spark_session()._jvm
        return self._jvm_view


_jvm_root = JvmPackage()


class Jvm:
    @classmethod
    def java_class(cls, name: str) -> JavaClass:  # noqa: ANN102
        java_class = getattr(_jvm_root, name)
        if isinstance(java_class, JavaClass):
            return java_class
        else:
            raise Exception(f"Class {name} not found in JVM.") from None



from .dp import Dp
from .dp_config import DpConfig
from .dp_result import DpResult
from .dp_type_inference_config import DpTypeInferenceConfig

__all__ = [
    "Dp",
    "DpConfig",
    "DpResult",
    "DpTypeInferenceConfig",
]



from typing import Dict, List, Optional

from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql.types import DataType

from ..spark.aggsum import AggSum, AggSumConfig, AggSumResult
from ..spark.dp import Dp, DpConfig, DpResult
from ..spark.dq import Dq, DqConfig, DqField, DqJoin, DqJoinResult, DqResult, DqRule
from ..spark.reshape import Reshape
from ..spark.reshape_config import ReshapeConfig


# TODO: Can we remove the duplication here and StoysApiPandas with some super class?
class PysparkDataFrameExtension:
    def __init__(self, sdf: SparkDataFrame) -> None:
        self._sdf = sdf

    def reshape(
        self,
        target_schema: DataType,
        config: Optional[ReshapeConfig] = None,
    ) -> SparkDataFrame:
        return Reshape.reshapeToDF(self._sdf, target_schema, config)

    def aggsum(
        self,
        key_column_names: Optional[List[str]] = None,
        referenced_table_names: Optional[List[str]] = None,
        config: Optional[AggSumConfig] = None,
    ) -> AggSumResult:
        aggsum = AggSum.fromDataFrame(self._sdf).config(config)
        aggsum.keyColumnNames(key_column_names).referencedTableNames(referenced_table_names)
        return aggsum.aggSumResult()

    def dp(
        self,
        config: Optional[DpConfig] = None,
    ) -> DpResult:
        dp = Dp.fromDataFrame(self._sdf).config(config)
        return dp.dpResult()

    def dq(
        self,
        rules: Optional[List[DqRule]] = None,
        fields: Optional[List[DqField]] = None,
        primary_key_field_names: Optional[List[str]] = None,
        metadata: Optional[Dict[str, str]] = None,
        config: Optional[DqConfig] = None,
    ) -> DqResult:
        dq = Dq.fromDataFrame(self._sdf).rules(rules).fields(fields)
        dq.primaryKeyFieldNames(primary_key_field_names).metadata(metadata).config(config)
        return dq.dqResult()

    def dq_equi_join(
        self,
        right_sdf: SparkDataFrame,
        left_column_names: List[str],
        right_column_names: List[str],
        metadata: Optional[Dict[str, str]] = None,
        config: Optional[DqConfig] = None,
    ) -> DqJoinResult:
        dq_join = DqJoin.equiJoin(self._sdf, right_sdf, left_column_names, right_column_names)
        dq_join.metadata(metadata).config(config)
        return dq_join.dqJoinResult()


def init() -> None:
    SparkDataFrame.stoys = property(lambda sdf: PysparkDataFrameExtension(sdf))



from typing import List

from ...ui.dq_join_result_ui import dq_join_result_to_html
from ...utils.case_class import CaseClassMirror, case_class_mirror
from .dq_result import DqResult


@case_class_mirror("io.stoys.spark.dq.DqJoinInfo")
class DqJoinInfo(CaseClassMirror):
    left_table_name: str
    right_table_name: str
    left_key_column_names: List[str]
    right_key_column_names: List[str]
    join_type: str
    join_condition: str


@case_class_mirror("io.stoys.spark.dq.DqJoinStatistics")
class DqJoinStatistics(CaseClassMirror):
    left_rows: int
    right_rows: int
    left_nulls: int
    right_nulls: int
    left_distinct: int
    right_distinct: int
    inner: int
    left: int
    right: int
    full: int
    cross: int


@case_class_mirror("io.stoys.spark.dq.DqJoinResult")
class DqJoinResult(CaseClassMirror):
    key: str
    dq_join_info: DqJoinInfo
    dq_join_statistics: DqJoinStatistics
    dq_result: DqResult

    def _repr_html_(self) -> str:
        return dq_join_result_to_html(self)



from typing import Union

from py4j.java_gateway import JavaClass, JavaObject, get_java_class

from ..utils.jvm import JvmPackage
from ..utils.py4j import Py4j

_jvm_scala = JvmPackage("io.stoys.scala")


class Reflection:
    @staticmethod
    def copy_case_class(case_class_instance: JavaObject, **kwargs) -> JavaObject:  # noqa: ANN003
        return _jvm_scala.Reflection.copyCaseClass(case_class_instance, Py4j.toMap(kwargs))

    @staticmethod
    def _class_name_to_type_tag(java_class: Union[JavaClass, str]) -> JavaObject:
        if isinstance(java_class, JavaClass):
            java_class_name = get_java_class(java_class).getName()
        elif isinstance(java_class, str):
            java_class_name = java_class
        else:
            raise Exception(f"Cannot get java class name from object of type {type(java_class)}") from None
        return _jvm_scala.Reflection.classNameToTypeTag(java_class_name)



from typing import Any, Dict

from .utils import stoys_ui_component_html

_default_pmf_ui_config = dict(
    dataType="float",
    showAxes=True,
    showLogScale=False,
    color="red",
)


def pmf_to_html(*pmf, **kwargs: Dict[str, Any]) -> str:  # noqa: ANN002
    payload = dict(
        data=[pmf.pmf for pmf in pmf],
        config={**_default_pmf_ui_config, **kwargs},
    )
    return stoys_ui_component_html("PmfPlot", payload)



from __future__ import annotations

from typing import Dict, List

from py4j.java_gateway import JavaObject
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import SQLContext
from pyspark.sql.column import Column

from ...utils.jvm import JvmPackage
from ...utils.py4j import Py4j
from ...utils.spark import Spark
from .dq_config import DqConfig
from .dq_join_result import DqJoinInfo, DqJoinResult
from .dq_join_type import DqJoinType

_jvm_dq = JvmPackage("io.stoys.spark.dq")


class DqJoin:
    def __init__(self, sql_ctx: SQLContext, java_dq_join: JavaObject) -> None:
        self._sql_ctx = sql_ctx
        self._java_dq_join = java_dq_join

    def config(self, config: DqConfig) -> DqJoin:
        if config:
            self._java_dq_join.config(config.to_jvm())
        return self

    def metadata(self, metadata: Dict[str, str]) -> DqJoin:
        if metadata:
            self._java_dq_join.metadata(Py4j.toMap(metadata))
        return self

    def joinType(self, join_type: DqJoinType) -> DqJoin:
        if join_type:
            jvm_join_type = getattr(_jvm_dq.DqJoinType, join_type.value)
            self._java_dq_join.joinType(jvm_join_type)
        return self

    def getDqJoinInfo(self) -> DqJoinInfo:
        return DqJoinInfo.from_jvm(self._java_dq_join.getDqJoinInfo())

    def computeDqJoinStatistics(self) -> SparkDataFrame:
        return SparkDataFrame(self._java_dq_join.computeDqJoinStatistics(), self._sql_ctx)

    def computeDqResult(self) -> SparkDataFrame:
        return SparkDataFrame(self._java_dq_join.computeDqResult(), self._sql_ctx)

    def computeDqJoinResult(self) -> SparkDataFrame:
        return SparkDataFrame(self._java_dq_join.computeDqJoinResult(), self._sql_ctx)

    def dqJoinResult(self) -> DqJoinResult:
        return DqJoinResult.from_jvm(self._java_dq_join.computeDqJoinResult().first())

    @staticmethod
    def equiJoin(
        left_sdf: SparkDataFrame,
        right_sdf: SparkDataFrame,
        left_column_names: List[str],
        right_column_names: List[str],
    ) -> DqJoin:
        java_dq_join = _jvm_dq.DqJoin.equiJoin(
            left_sdf._jdf,
            right_sdf._jdf,
            Py4j.toSeq(left_column_names),
            Py4j.toSeq(right_column_names),
            Spark.rowTypeTag,
            Spark.rowTypeTag,
        )
        return DqJoin(left_sdf.sql_ctx, java_dq_join)

    @staticmethod
    def expensiveArbitraryJoin(left_sdf: SparkDataFrame, right_sdf: SparkDataFrame, join_condition: Column) -> DqJoin:
        java_dq_join = _jvm_dq.DqJoin.expensiveArbitraryJoin(
            left_sdf._jdf,
            right_sdf._jdf,
            join_condition._jc,
            Spark.rowTypeTag,
            Spark.rowTypeTag,
        )
        return DqJoin(left_sdf.sql_ctx, java_dq_join)



from py4j.java_gateway import JavaClass
from pyspark.sql import SparkSession

from ..config import STOYS_LIB_MIN_JAVA_VERSION

_is_spark_session_builder_patched: bool = False


def _patch_spark_session_builder() -> None:
    global _is_spark_session_builder_patched
    if not _is_spark_session_builder_patched:
        from ..deps.use import add_stoys_jars

        original_spark_session_builder_get_or_create = SparkSession.Builder.getOrCreate
        SparkSession.Builder.getOrCreate = lambda b: original_spark_session_builder_get_or_create(add_stoys_jars(b))
        _is_spark_session_builder_patched = True


def get_or_create_spark_session() -> SparkSession:
    spark_sesion = SparkSession.getActiveSession()
    if spark_sesion is None:
        _patch_spark_session_builder()
        spark_sesion = SparkSession.builder.getOrCreate()
        check_spark_session_with_stoys_library()
    return spark_sesion


def check_spark_session_with_stoys_library() -> None:
    spark_sesion = SparkSession.getActiveSession()
    if spark_sesion is None:
        raise Exception("SparkSession.getActiveSession() is None. Have you created SparkSession yet?")
    jvm_view = spark_sesion._jvm
    java_version = jvm_view.java.lang.System.getProperty("java.version")
    if java_version < STOYS_LIB_MIN_JAVA_VERSION:
        raise Exception(
            f"""Current Java version is {java_version}.
                Minimum required Java version is {STOYS_LIB_MIN_JAVA_VERSION}.
                Consider setting JAVA_HOME to newer Java before starting SparkSession."""  # noqa: C812
        )
    if not isinstance(jvm_view.io.stoys.scala.Jackson, JavaClass):
        # TODO: Show nice instructions how to add our jars to the environment.
        #       Note: This is very different for different environments. jupyterlab / databricks / zeppelin
        raise Exception("TODO: Missing stoys jars! Please add them to spark and restart SparkSession.")



from typing import Any, Dict, Optional

import pandas as pd
from IPython.core import magic
from IPython.core.getipython import get_ipython
from IPython.core.magic_arguments import argument, magic_arguments, parse_argstring
from pyspark.sql import DataFrame as SparkDataFrame

from ..spark.aggsum import AggSum, AggSumResult
from ..spark.dp import Dp, DpResult
from ..spark.dq import Dq, DqResult
from ..utils.spark_session import get_or_create_spark_session


@magic.magics_class
class IpythonMagics(magic.Magics):
    _table_name_help = """
        Table name is used to register the result as Spark (temporary) view and Pyhton variable.
        BEWARE: It will overwrite any existing view / variable of the same name.
    """.strip()

    @magic.cell_magic
    @magic.needs_local_scope
    @magic_arguments()
    @argument("-dl", "--display_limit", type=int, nargs="?", default=1_000, help="Maximum number of rows to display.")
    @argument("table_name", type=str, nargs="?", help=_table_name_help)
    def ssql(self, line: str, cell: str, local_ns: Dict[str, Any]) -> pd.DataFrame:
        """Execute Spark SQL command and display the result."""
        args = parse_argstring(self.ssql, line)
        sdf = self._execute_sql(cell, local_ns, args.table_name)
        sdf._repr_html_ = lambda: self._safe_limit(sdf, args.display_limit).toPandas()._repr_html_()
        return sdf

    @magic.cell_magic
    @magic.needs_local_scope
    def ssql_dq(self, line: str, cell: str, local_ns: Dict[str, Any]) -> DqResult:
        """Compute DQ (quality) from SQL and display the result."""
        sql = cell.format(**local_ns)
        return Dq.fromDqSql(get_or_create_spark_session(), sql).dqResult()

    @magic.cell_magic
    @magic.needs_local_scope
    def ssql_dp(self, line: str, cell: str, local_ns: Dict[str, Any]) -> DpResult:
        """Compute DP (profile) from SQL and display the result."""
        sdf = self._execute_sql(cell, local_ns, table_name=None)
        return Dp.fromDataFrame(sdf).dpResult()

    @magic.cell_magic
    @magic.needs_local_scope
    def ssql_aggsum(self, line: str, cell: str, local_ns: Dict[str, Any]) -> AggSumResult:
        """Compute AggSum (aggregate summary) from SQL and display the result."""
        sdf = self._execute_sql(cell, local_ns, table_name=None)
        return AggSum.fromDataFrame(sdf).aggSumResult()

    def _execute_sql(self, sql: str, local_ns: Dict[str, Any], table_name: Optional[str]) -> SparkDataFrame:
        sql = sql.format(**local_ns)
        sdf = get_or_create_spark_session().sql(sql)
        if table_name is not None:
            sdf = sdf.alias(table_name)
            sdf.createOrReplaceTempView(table_name)
            # globals()[args.table_name] = sdf
            get_ipython().push({table_name: sdf})
        return sdf

    def _safe_limit(self, sdf: SparkDataFrame, limit: int) -> SparkDataFrame:
        return sdf if limit is None or limit <= 0 else sdf.limit(limit)


def init() -> None:
    get_ipython().register_magics(IpythonMagics)



from dataclasses import dataclass, replace
from typing import ClassVar, Type, TypeVar

import serde

# from dataclasses_json import dataclass_json
from py4j.java_gateway import JavaClass, JavaObject

from ..scala import Jackson
from .json import Json
from .jvm import Jvm
from .python import classproperty

CCM = TypeVar("CCM", bound="CaseClassMirror")


class CaseClassMirror(object):
    _stoys_: ClassVar[str]
    _java_class: ClassVar[JavaClass] = None

    @classproperty
    def java_class(cls: Type[CCM]) -> JavaClass:  # noqa: B902,N805
        if cls._java_class is None:
            cls._java_class = Jvm.java_class(cls._stoys_)
        return cls._java_class

    def copy(self: CCM, **kwargs) -> CCM:  # noqa: ANN003
        return replace(self, **kwargs)

    def to_json(self: CCM) -> str:
        return Json.to_json(self)

    @classmethod
    def from_json(cls: Type[CCM], json: str) -> CCM:
        return Json.from_json(cls, json)

    def to_jvm(self: CCM) -> JavaObject:
        return Jackson.fromJson(self.java_class, self.to_json())

    @classmethod
    def from_jvm(cls: Type[CCM], java_object: JavaObject) -> CCM:
        return cls.from_json(Jackson.toJson(java_object))
        # return cls.from_json(Jackson.toJson(java_object), infer_missing=True)


def case_class_mirror(jvm_class_name: str) -> CCM:
    def wrapper(cls: Type[CCM]) -> CCM:
        cls = serde.deserialize()(serde.serialize()(dataclass(frozen=True)(cls)))
        # cls = dataclass_json()(dataclass(frozen=True)(cls))
        cls._stoys_ = jvm_class_name
        return cls

    return wrapper



import importlib
from enum import Enum
from typing import Any


def is_module_installed(module_name: str) -> bool:
    return importlib.util.find_spec(module_name) is not None


class classproperty(property):  # noqa: N801
    def __get__(self, cls, owner) -> Any:  # noqa: ANN001
        return classmethod(self.fget).__get__(None, owner)()


class AutoNameEnum(Enum):
    def _generate_next_value_(name, start, count, last_values) -> str:  # noqa: ANN001, B902,N805
        return name



from .use import (
    add_stoys_jars,
    add_stoys_jars_packages,
    get_stoys_jars,
    get_stoys_jars_packages,
    get_stoys_ui_js_content,
)

__all__ = [
    "get_stoys_jars_packages",
    "get_stoys_jars",
    "get_stoys_ui_js_content",
    "add_stoys_jars",
    "add_stoys_jars_packages",
]



from typing import Optional

from pyspark.sql import DataFrame as SparkDataFrame

from ..utils.jvm import JvmPackage

_jvm_spark = JvmPackage("io.stoys.spark")


class Datasets:
    @staticmethod
    def getAlias(sdf: SparkDataFrame) -> Optional[str]:
        return _jvm_spark.Datasets.getAlias(sdf._jdf)



from __future__ import annotations

from typing import List, Optional

from pyspark.sql.types import DataType

from ...utils.data_types import DataTypes
from ...utils.jvm import JvmPackage
from ...utils.py4j import Py4j
from .dq_field import DqField
from .dq_rule import DqRule

_jvm_dq = JvmPackage("io.stoys.spark.dq")


class DqRules:
    @classmethod
    def name(cls: DqRules, field_name: str, logical_name: str) -> str:
        # return f"{field_name}{LOGICAL_NAME_SEPARATOR}{logical_name}"
        return _jvm_dq.DqRules.name(field_name, logical_name)

    @classmethod
    def namedRule(
        cls: DqRules,
        field_name: str,
        logical_name: str,
        expression: str,
        description: Optional[str] = None,
    ) -> DqRule:
        # return DqRules.rule(DqRules.name(field_name, logical_name), expression, description)
        return DqRule.from_jvm(_jvm_dq.DqRules.namedRule(field_name, logical_name, expression, description))

    @classmethod
    def nullSafeNamedRule(
        cls: DqRules,
        field_name: str,
        logical_name: str,
        expression: str,
        description: Optional[str] = None,
    ) -> DqRule:
        # return DqRules.named_rule(
        #     field_name, logical_name, f"{SqlUtils.quoteIfNeeded(field_name)} IS NULL OR ({expression})", description
        # )
        return DqRule.from_jvm(_jvm_dq.DqRules.nullSafeNamedRule(field_name, logical_name, expression, description))

    # TODO: match all rule(...) variants from scala class
    @classmethod
    def rule(
        cls: DqRules,
        name: str,
        expression: str,
        description: Optional[str] = None,
    ) -> DqRule:
        # return DqRule(name, expression, description, [])
        return DqRule.from_jvm(_jvm_dq.DqRules.rule(name, expression, description))

    @classmethod
    def field(
        cls: DqRules,
        name: str,
        data_type_json: str,
        nullable: bool = True,
        enum_values: Optional[List[str]] = None,
        format: Optional[str] = None,
        regexp: Optional[str] = None,
    ) -> DqRule:
        jvm_enum_values = Py4j.toSeq(enum_values or [])
        return DqField.from_jvm(_jvm_dq.DqRules.field(name, data_type_json, nullable, jvm_enum_values, format, regexp))

    # common rules

    @classmethod
    def enumValuesRule(
        cls: DqRules,
        field_name: str,
        enum_values: Optional[List[str]] = None,
        case_insensitive: bool = True,
    ) -> DqRule:
        jvm_enum_values = Py4j.toSeq(enum_values or [])
        return DqRule.from_jvm(_jvm_dq.DqRules.enumValuesRule(field_name, jvm_enum_values, case_insensitive))

    @classmethod
    def notNullRule(
        cls: DqRules,
        field_name: str,
    ) -> DqRule:
        return DqRule.from_jvm(_jvm_dq.DqRules.notNullRule(field_name))

    @classmethod
    def regexpRule(
        cls: DqRules,
        field_name: str,
        regexp: Optional[str] = None,
    ) -> DqRule:
        return DqRule.from_jvm(_jvm_dq.DqRules.regexpRule(field_name, regexp))

    @classmethod
    def typeRule(
        cls: DqRules,
        field_name: str,
        source_type: DataType,
        target_type: DataType,
        format: Optional[str] = None,
    ) -> DqRule:
        jvm_source_type = DataTypes.to_jvm(source_type)
        jvm_target_type = DataTypes.to_jvm(target_type)
        return DqRule.from_jvm(_jvm_dq.DqRules.typeRule(field_name, jvm_source_type, jvm_target_type, format))

    # composite and multi field rules

    @classmethod
    def all(
        cls: DqRules,
        rule_name: str,
        rules: List[DqRule],
        description: Optional[str] = None,
    ) -> DqRule:
        return DqRule.from_jvm(_jvm_dq.DqRules.all(rule_name), Py4j.toSeq(r.to_jvm() for r in rules), description)

    @classmethod
    def any(
        cls: DqRules,
        rule_name: str,
        rules: List[DqRule],
        description: Optional[str] = None,
    ) -> DqRule:
        return DqRule.from_jvm(_jvm_dq.DqRules.any(rule_name), Py4j.toSeq(r.to_jvm() for r in rules), description)

    @classmethod
    def uniqueRule(
        cls: DqRules,
        base_rule_name: str,
        field_names: Optional[List[str]] = None,
    ) -> DqRule:
        if field_names is None:
            return DqRule.from_jvm(_jvm_dq.DqRules.uniqueRule(base_rule_name))
        else:
            return DqRule.from_jvm(_jvm_dq.DqRules.uniqueRule(base_rule_name, Py4j.toSeq(field_names)))



from typing import Union

from py4j.java_gateway import JavaClass, JavaObject

from ..utils.jvm import JvmPackage
from .reflection import Reflection

_jvm_scala = JvmPackage("io.stoys.scala")


class Arbitrary:
    @staticmethod
    def default(java_class: Union[JavaClass, str]) -> JavaObject:
        return _jvm_scala.Arbitrary.default(Reflection._class_name_to_type_tag(java_class))

    @staticmethod
    def empty(java_class: Union[JavaClass, str]) -> JavaObject:
        return _jvm_scala.Arbitrary.empty(Reflection._class_name_to_type_tag(java_class))

    @staticmethod
    def hashed(java_class: Union[JavaClass, str], salt: int) -> JavaObject:
        return _jvm_scala.Arbitrary.hashed(salt, Reflection._class_name_to_type_tag(java_class))

    @staticmethod
    def indexed(java_class: Union[JavaClass, str], salt: int) -> JavaObject:
        return _jvm_scala.Arbitrary.indexed(salt, Reflection._class_name_to_type_tag(java_class))

    @staticmethod
    def proto(java_class: Union[JavaClass, str]) -> JavaObject:
        return _jvm_scala.Arbitrary.proto(Reflection._class_name_to_type_tag(java_class))

    @staticmethod
    def random(java_class: Union[JavaClass, str], seed: int) -> JavaObject:
        return _jvm_scala.Arbitrary.random(seed, Reflection._class_name_to_type_tag(java_class))



import pathlib

from ..config import STOYS_EMBEDDED_DEPENDENCIES_PACKAGE

_repository_dir = pathlib.Path(__file__).parent.parent.parent.resolve()
embedded_dependencies_dir = _repository_dir.joinpath(*STOYS_EMBEDDED_DEPENDENCIES_PACKAGE.split("."))



from __future__ import annotations

from typing import Dict, List, Optional

from ..utils.case_class import CaseClassMirror, case_class_mirror
from ..utils.python import classproperty
from .reshape import ReshapeConfig


@case_class_mirror("io.stoys.spark.SparkIOConfig")
class SparkIOConfig(CaseClassMirror):
    input_paths: List[str]
    input_reshape_config: ReshapeConfig
    register_input_tables: bool
    output_path: Optional[str]
    write_format: Optional[str]
    write_mode: Optional[str]
    write_options: Dict[str, str]

    @classproperty
    def default(cls) -> SparkIOConfig:  # noqa: B902, N805
        return cls.from_jvm(cls.java_class.default())

    @classproperty
    def notebook(cls) -> SparkIOConfig:  # noqa: B902, N805
        return cls.from_jvm(cls.java_class.notebook())



from __future__ import annotations

from typing import List

from ...utils.case_class import CaseClassMirror, case_class_mirror
from ...utils.python import classproperty


@case_class_mirror("io.stoys.spark.dp.DpTypeInferenceConfig.EnumValues")
class EnumValues(CaseClassMirror):
    name: str
    values: List[str]


@case_class_mirror("io.stoys.spark.dp.DpTypeInferenceConfig")
class DpTypeInferenceConfig(CaseClassMirror):
    prefer_not_nullable: bool
    prefer_boolean_for_01: bool
    prefer_float: bool
    prefer_integer: bool
    date_formats: List[str]
    timestamp_formats: List[str]
    enum_values: List[EnumValues]

    @classproperty
    def default(cls) -> DpTypeInferenceConfig:  # noqa: B902, N805
        return cls.from_jvm(cls.java_class.default())



from typing import List, Optional

from ...utils.case_class import CaseClassMirror, case_class_mirror


@case_class_mirror("io.stoys.spark.dq.DqRule")
class DqRule(CaseClassMirror):
    name: str
    expression: str
    description: Optional[str]
    referenced_column_names: List[str]



from __future__ import annotations

from typing import List

from py4j.java_gateway import JavaObject
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import SQLContext
from pyspark.sql.session import SparkSession

from ...utils.jvm import JvmPackage
from ...utils.py4j import Py4j
from .aggsum_config import AggSumConfig
from .aggsum_result import AggSumInfo, AggSumResult

_jvm_aggsum = JvmPackage("io.stoys.spark.aggsum")


class AggSum:
    def __init__(self, sql_ctx: SQLContext, java_aggsum: JavaObject) -> None:
        self._sql_ctx = sql_ctx
        self._java_aggsum = java_aggsum

    def config(self, config: AggSumConfig) -> AggSum:
        if config:
            self._java_aggsum.config(config.to_jvm())
        return self

    def keyColumnNames(self, key_column_names: List[str]) -> AggSum:
        if key_column_names:
            self._java_aggsum.keyColumnNames(Py4j.toSeq(key_column_names))
        return self

    def referencedTableNames(self, referenced_table_names: List[str]) -> AggSum:
        if referenced_table_names:
            self._java_aggsum.referencedTableNames(Py4j.toSeq(referenced_table_names))
        return self

    def getAggSumInfo(self) -> AggSumInfo:
        return AggSumInfo.from_jvm(self._java_aggsum.getAggSumInfo())

    def computeAggSum(self) -> SparkDataFrame:
        return SparkDataFrame(self._java_aggsum.computeAggSum(), self._sql_ctx)

    def computeAggSumResult(self) -> SparkDataFrame:
        return SparkDataFrame(self._java_aggsum.computeAggSumResult(), self._sql_ctx)

    def aggSumResult(self) -> AggSumResult:
        return AggSumResult.from_jvm(self._java_aggsum.computeAggSumResult().first())

    @staticmethod
    def fromDataFrame(sdf: SparkDataFrame) -> AggSum:
        return AggSum(sdf.sql_ctx, _jvm_aggsum.AggSum.fromDataFrame(sdf._jdf))

    # @staticmethod
    # def fromDataset(sdf: SparkDataFrame) -> AggSum:
    #     return AggSum(sdf.sql_ctx, _jvm_aggsum.AggSum.fromDataset(sdf._jdf))

    @staticmethod
    def fromSql(spark_session: SparkSession, sql: str) -> AggSum:
        sql_ctx = SQLContext(spark_session.sparkContext)
        return AggSum(sql_ctx, _jvm_aggsum.AggSum.fromSql(spark_session._jsparkSession, sql))



from typing import Any

# from typeguard.importhook import install_import_hook

# install_import_hook("stoys")


def init() -> Any:
    from .utils.python import is_module_installed

    if is_module_installed("pyspark"):
        from pyspark.sql import SparkSession

        if SparkSession.getActiveSession() is None:
            from .utils.spark_session import _patch_spark_session_builder

            _patch_spark_session_builder()
        else:
            from .utils.spark_session import check_spark_session_with_stoys_library

            check_spark_session_with_stoys_library()
        from .integration import pyspark as pyspark_integration

        pyspark_integration.init()
    else:
        raise Exception("Stoys cannot work without spark! Please install package pyspark!") from None

    if is_module_installed("pandas"):
        from .integration import pandas as pandas_integration

        pandas_integration.init()

    if is_module_installed("IPython"):
        from .integration import ipython as ipython_integration

        ipython_integration.init()

    # TODO: koalas
    # TODO: pyspark.pandas

    from .deps.use import get_stoys_ui_js_content
    from .integration.notebook import display_html
    from .ui.utils import wrap_javascript_in_html

    return display_html(wrap_javascript_in_html(get_stoys_ui_js_content()))



import uuid
from typing import Any

from ..utils.json import Json

DEFAULT_HEIGHT: str = "24rem"


def wrap_javascript_in_html(javascript: str) -> str:
    element_id = f"stoys-element-{uuid.uuid4()}"
    return f"""
        <span id="{element_id}"></span>
        <script>
            try {{
                let element = document.getElementById("{element_id}");
                {javascript}
            }} catch (error) {{
                let element = document.getElementById("{element_id}");
                element.innerHTML = "JavascriptError: " + error;
            }}
        </script>
    """


def stoys_ui_component_html(component_name: str, payload: Any) -> str:
    return wrap_javascript_in_html(
        f"""
        (function() {{
            const {{{component_name}, ReactDOM, React}} = stoysUi;
            ReactDOM.render(React.createElement({component_name}, {Json.to_json(payload)}, null), element);
        }})()
        """  # noqa: C812
    )



from typing import List

STOYS_UI_VERSION: str = "0.2.0"

STOYS_LIB_MIN_JAVA_VERSION: str = "1.8"
STOYS_LIB_VERSION: str = "0.5.0-SNAPSHOT"
STOYS_LIB_GROUP_ID: str = "io.stoys"
STOYS_LIB_MODULES: List[str] = ["stoys-scala", "stoys-spark", "stoys-spark-du"]
STOYS_LIB_SCALA_COMPAT_VERSION: str = "2.12"
STOYS_LIB_CLASSIFIER: str = "spark-shaded"

STOYS_EMBEDDED_DEPENDENCIES_PACKAGE: str = "stoys.deps._embedded"



from typing import List, Optional

from pyspark.sql import SparkSession

from ..config import (
    STOYS_LIB_CLASSIFIER,
    STOYS_LIB_GROUP_ID,
    STOYS_LIB_MODULES,
    STOYS_LIB_SCALA_COMPAT_VERSION,
    STOYS_LIB_VERSION,
)
from .common import embedded_dependencies_dir


def get_stoys_jars_packages() -> List[str]:
    def get_module_coords(module: str) -> str:
        return (
            f"{STOYS_LIB_GROUP_ID}:{module}_{STOYS_LIB_SCALA_COMPAT_VERSION}:{STOYS_LIB_VERSION}:{STOYS_LIB_CLASSIFIER}"
        )

    return [get_module_coords(module) for module in STOYS_LIB_MODULES]


def get_stoys_jars() -> List[str]:
    # return [ str(p) for p in embedded_dependencies_dir.glob("*.jar") ]
    def get_module_file_name(module: str) -> str:
        return f"{module}_{STOYS_LIB_SCALA_COMPAT_VERSION}-{STOYS_LIB_VERSION}-{STOYS_LIB_CLASSIFIER}.jar"

    return [str(embedded_dependencies_dir.joinpath(get_module_file_name(module))) for module in STOYS_LIB_MODULES]


def get_stoys_ui_js_content() -> str:
    return embedded_dependencies_dir.joinpath("stoys-ui.js").read_text(encoding="UTF8")


def _append_value_list(separator: str, current_value_list: Optional[str], *values: str) -> str:
    current_values = current_value_list.split(separator) if current_value_list is not None else []
    all_values = current_values + [v for v in values if v is not None and v not in current_values]
    return separator.join(all_values)


def add_stoys_jars(builder: SparkSession.Builder) -> SparkSession.Builder:
    current_jars = builder._options.get("spark.jars")
    jars = _append_value_list(",", current_jars, *get_stoys_jars())
    builder.config("spark.jars", jars)
    return builder


def add_stoys_jars_packages(builder: SparkSession.Builder) -> SparkSession.Builder:
    current_packages = builder._options.get("spark.jars.packages")
    packages = _append_value_list(",", current_packages, *get_stoys_jars_packages())
    builder.config("spark.jars.packages", packages)
    return builder



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['stoys',
 'stoys.deps',
 'stoys.integration',
 'stoys.scala',
 'stoys.spark',
 'stoys.spark.aggsum',
 'stoys.spark.dp',
 'stoys.spark.dq',
 'stoys.ui',
 'stoys.utils']

package_data = \
{'': ['*'], 'stoys.deps': ['_embedded/*']}

install_requires = \
['pandas>=1,<2', 'pyserde>=0.5.0', 'pyspark>=3,<3.3']

entry_points = \
{'console_scripts': ['get-deps = stoys.deps.download:download_dependencies']}

setup_kwargs = {
    'name': 'stoys',
    'version': '0.4.6',
    'description': 'Stoys: Spark Tools @ stoys.io',
    'long_description': None,
    'author': 'Stoys Authors',
    'author_email': 'authors@stoys.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/stoys-io/stoys-python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4',
}


setup(**setup_kwargs)



from typing import Dict, List, Optional

import pandas as pd
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql.types import DataType

from ..spark.aggsum import AggSum, AggSumConfig, AggSumResult
from ..spark.dp import Dp, DpConfig, DpResult
from ..spark.dq import Dq, DqConfig, DqField, DqJoin, DqJoinResult, DqResult, DqRule
from ..spark.reshape import Reshape
from ..spark.reshape_config import ReshapeConfig
from ..utils.spark_session import get_or_create_spark_session


@pd.api.extensions.register_dataframe_accessor("stoys")
class PandasDataFrameExtension:
    def __init__(self, df: pd.DataFrame) -> None:
        self._df = df
        self._sdf = None

    def _to_spark_data_frame(self, df: pd.DataFrame) -> SparkDataFrame:
        return get_or_create_spark_session().createDataFrame(df)

    def _get_sdf(self) -> SparkDataFrame:
        if self._sdf is None:
            self._sdf = self._to_spark_data_frame(self._df)
        return self._sdf

    def reshape(
        self,
        target_schema: DataType,
        config: Optional[ReshapeConfig] = None,
    ) -> pd.DataFrame:
        return Reshape.reshapeToDF(self._get_sdf(), target_schema, config).toPandas()

    def aggsum(
        self,
        key_column_names: Optional[List[str]] = None,
        referenced_table_names: Optional[List[str]] = None,
        config: Optional[AggSumConfig] = None,
    ) -> AggSumResult:
        aggsum = AggSum.fromDataFrame(self._get_sdf()).config(config)
        aggsum.keyColumnNames(key_column_names or list(self._df.index.names))
        aggsum.referencedTableNames(referenced_table_names)
        return aggsum.aggSumResult()

    def dp(
        self,
        config: Optional[DpConfig] = None,
    ) -> DpResult:
        dp = Dp.fromDataFrame(self._get_sdf()).config(config)
        return dp.dpResult()

    def dq(
        self,
        rules: Optional[List[DqRule]] = None,
        fields: Optional[List[DqField]] = None,
        primary_key_field_names: Optional[List[str]] = None,
        metadata: Optional[Dict[str, str]] = None,
        config: Optional[DqConfig] = None,
    ) -> DqResult:
        dq = Dq.fromDataFrame(self._get_sdf()).rules(rules).fields(fields)
        dq.primaryKeyFieldNames(primary_key_field_names).metadata(metadata).config(config)
        return dq.dqResult()

    def dq_equi_join(
        self,
        right_df: pd.DataFrame,
        left_column_names: List[str],
        right_column_names: List[str],
        metadata: Optional[Dict[str, str]] = None,
        config: Optional[DqConfig] = None,
    ) -> DqJoinResult:
        right_sdf = self._to_spark_data_frame(right_df)
        dq_join = DqJoin.equiJoin(self._get_sdf(), right_sdf, left_column_names, right_column_names)
        dq_join.metadata(metadata).config(config)
        return dq_join.dqJoinResult()


def init() -> None:
    pass



from .datasets import Datasets
from .reshape import Reshape
from .reshape_config import ReshapeConfig
from .spark_io import SparkIO, SparkIOContext
from .spark_io_config import SparkIOConfig
from .table_name import TableName

__all__ = [
    "Datasets",
    "Reshape",
    "ReshapeConfig",
    "SparkIO",
    "SparkIOConfig",
    "SparkIOContext",
    "TableName",
]



from .dq import Dq
from .dq_config import DqConfig
from .dq_field import DqField
from .dq_join import DqJoin
from .dq_join_result import DqJoinResult
from .dq_join_type import DqJoinType
from .dq_result import DqResult
from .dq_rule import DqRule
from .dq_rules import DqRules

__all__ = [
    "Dq",
    "DqConfig",
    "DqField",
    "DqJoin",
    "DqJoinResult",
    "DqJoinType",
    "DqResult",
    "DqRule",
    "DqRules",
    "DqRules",
]



import builtins
from typing import Any, Callable, Optional

from ..utils.python import is_module_installed


def is_ipython() -> bool:
    return getattr(builtins, "__IPYTHON__", False) and is_module_installed("IPython.core.display")


def is_databricks() -> bool:
    return is_ipython() and hasattr(globals(), "dbutils") and hasattr(globals(), "displayHTML")


def is_zeppelin() -> bool:
    return "ZeppelinContext" in str(type(getattr(globals(), "z", None)))


def is_colab() -> bool:
    if not is_ipython():
        return False
    else:
        from IPython.core.getipython import get_ipython

        return is_module_installed("google.colab") and "google.colab" in str(get_ipython())


_display_html_function: Optional[Callable[[str], Any]] = None


def _display_html_ipython(html: str) -> None:
    from IPython.core import display

    display.display(display.HTML(html))


def _display_html_databricks(html: str) -> None:
    global displayHTML
    displayHTML(html)  # type: ignore


def _display_html_zeppelin(html: str) -> None:
    print(f"%html {html}")


def _display_html_unsupported(html: str) -> None:
    print(f"No notebook environment detected! HTML snipped of length {len(html)} is ignored.")


def _detect_display_html_function() -> Optional[Callable[[str], Any]]:
    if is_ipython():
        if is_databricks():
            return _display_html_databricks
        return _display_html_ipython
    else:
        if is_zeppelin():
            return _display_html_zeppelin
    return _display_html_unsupported


def display_html(html: str) -> Any:
    global _display_html_function
    if _display_html_function is None:
        _display_html_function = _detect_display_html_function()
    return _display_html_function(html)



from typing import Any, Dict

from .utils import stoys_ui_component_html

_default_dq_result_ui_config = dict(
    # mode="row",
    # mode="column",
    smallSize=True,
    # selectedRules=[],
    pagination=False,
    # showReferencedColumnsOnly=False,
    # rulesTableProps={},
    # sampleTableProps={},
)


def dq_result_to_html(dq_result, **kwargs: Dict[str, Any]) -> str:  # noqa: ANN001
    payload = dict(
        data=dq_result,
        config={**_default_dq_result_ui_config, **kwargs},
    )
    return stoys_ui_component_html("Quality", payload)



from py4j.java_gateway import JavaObject

from .jvm import JvmPackage
from .python import classproperty

_jvm_utils = JvmPackage("io.stoys.utils")


class Spark:
    @classproperty
    def rowTypeTag(cls) -> JavaObject:  # noqa: B902,N805
        return _jvm_utils.Spark.rowTypeTag()



from typing import Any, Dict

from .utils import DEFAULT_HEIGHT, stoys_ui_component_html

_default_dp_result_ui_config = dict(
    # colors=["red", "green", "blue",],
    height=DEFAULT_HEIGHT,
    smallSize=True,
    # orientType="Horizontal",
    # orientType="Vertical",
    pagination=False,
    # visibleColumns=["count_nulls", "count_unique", ],
    # showAxesSwitcher=True,
    # showChartTableSwitcher=True,
    # showJsonSwitcher=True,
    # showLogarithmicSwitcher=True,
    # showNormalizeSwitcher=True,
    # showOrientSwitcher=True,
    # showProfilerToolbar=True,
    # showSearch=True,
    # axesChecked=True,
    # chartTableChecked=True,
    # jsonChecked=True,
    # normalizeChecked=True,
    # logarithmicChecked=True,
)


def dp_result_to_html(*dp_result, **kwargs: Dict[str, Any]) -> str:  # noqa: ANN002
    payload = dict(
        datasets=list(dp_result),
        config={**_default_dp_result_ui_config, **kwargs},
    )
    return stoys_ui_component_html("Profiler", payload)
